"""
vidformer.cv2 is the cv2 frontend for [vidformer](https://github.com/ixlab/vidformer).

**Quick links:**
* [📦 PyPI](https://pypi.org/project/vidformer/)
* [📘 Documentation - vidformer-py](https://ixlab.github.io/vidformer/vidformer-py/)
* [📘 Documentation - vidformer.cv2](https://ixlab.github.io/vidformer/vidformer-py/vidformer/cv2.html)
* [📘 Documentation - vidformer.supervision](https://ixlab.github.io/vidformer/vidformer-py/vidformer/supervision.html)
* [🧑‍💻 Source Code](https://github.com/ixlab/vidformer/tree/main/vidformer-py/)
"""

import vidformer as vf

try:
    import cv2 as _opencv2
except Exception:
    _opencv2 = None

import re
import zlib
from bisect import bisect_right
from fractions import Fraction
from urllib.parse import unquote
import os

import numpy as np

CAP_PROP_POS_MSEC = 0
CAP_PROP_POS_FRAMES = 1
CAP_PROP_FRAME_WIDTH = 3
CAP_PROP_FRAME_HEIGHT = 4
CAP_PROP_FPS = 5
CAP_PROP_FRAME_COUNT = 7

FONT_HERSHEY_SIMPLEX = 0
FONT_HERSHEY_PLAIN = 1
FONT_HERSHEY_DUPLEX = 2
FONT_HERSHEY_COMPLEX = 3
FONT_HERSHEY_TRIPLEX = 4
FONT_HERSHEY_COMPLEX_SMALL = 5
FONT_HERSHEY_SCRIPT_SIMPLEX = 6
FONT_HERSHEY_SCRIPT_COMPLEX = 7
FONT_ITALIC = 16

FILLED = -1
LINE_4 = 4
LINE_8 = 8
LINE_AA = 16

MARKER_CROSS = 0
MARKER_TILTED_CROSS = 1
MARKER_STAR = 2
MARKER_DIAMOND = 3
MARKER_SQUARE = 4
MARKER_TRIANGLE_UP = 5
MARKER_TRIANGLE_DOWN = 6

INTER_NEAREST = 0
INTER_LINEAR = 1
INTER_CUBIC = 2
INTER_AREA = 3
INTER_LANCOZOS4 = 4
INTER_LINEAR_EXACT = 5
INTER_NEAREST_EXACT = 6
INTER_MAX = 7

# Rotation flags
ROTATE_90_CLOCKWISE = 0
ROTATE_180 = 1
ROTATE_90_COUNTERCLOCKWISE = 2

# Border types
BORDER_CONSTANT = 0
BORDER_REPLICATE = 1
BORDER_REFLECT = 2
BORDER_WRAP = 3
BORDER_REFLECT_101 = 4
BORDER_TRANSPARENT = 5
BORDER_REFLECT101 = BORDER_REFLECT_101
BORDER_DEFAULT = BORDER_REFLECT_101
BORDER_ISOLATED = 16

_inline_mat = vf.Filter("_inline_mat")
_slice_mat = vf.Filter("_slice_mat")
_slice_write_mat = vf.Filter("_slice_write_mat")
_black = vf.Filter("_black")
_solid = vf.Filter("_solid")


_filter_scale = vf.Filter("Scale")
_filter_rectangle = vf.Filter("cv2.rectangle")
_filter_putText = vf.Filter("cv2.putText")
_filter_arrowedLine = vf.Filter("cv2.arrowedLine")
_filter_line = vf.Filter("cv2.line")
_filter_circle = vf.Filter("cv2.circle")
_filter_addWeighted = vf.Filter("cv2.addWeighted")
_filter_ellipse = vf.Filter("cv2.ellipse")
_filter_polylines = vf.Filter("cv2.polylines")
_filter_fillPoly = vf.Filter("cv2.fillPoly")
_filter_fillConvexPoly = vf.Filter("cv2.fillConvexPoly")
_filter_drawContours = vf.Filter("cv2.drawContours")
_filter_drawMarker = vf.Filter("cv2.drawMarker")
_set_to = vf.Filter("cv2.setTo")
_filter_flip = vf.Filter("cv2.flip")
_filter_rotate = vf.Filter("cv2.rotate")
_filter_copyMakeBorder = vf.Filter("cv2.copyMakeBorder")
_filter_hconcat = vf.Filter("cv2.hconcat")
_filter_vconcat = vf.Filter("cv2.vconcat")


def _ts_to_fps(timestamps):
    if len(timestamps) < 2:
        return 0
    fps = Fraction(len(timestamps), timestamps[-1] - timestamps[0])
    if fps.denominator == 1:
        return fps.numerator
    return float(fps)


def _fps_to_ts(fps, n_frames):
    assert type(fps) is int
    return [Fraction(i, fps) for i in range(n_frames)]


_global_cv2_server = None


def _server():
    global _global_cv2_server
    if _global_cv2_server is None:
        if "VF_IGNI_ENDPOINT" in os.environ:
            server_endpoint = os.environ["VF_IGNI_ENDPOINT"]
            if "VF_IGNI_API_KEY" not in os.environ:
                raise Exception("VF_IGNI_API_KEY must be set")
            api_key = os.environ["VF_IGNI_API_KEY"]
            _global_cv2_server = vf.Server(server_endpoint, api_key)
        else:
            raise Exception(
                "No server set for the cv2 frontend (https://ixlab.github.io/vidformer/docs/install.html). Set VF_IGNI_ENDPOINT and VF_IGNI_API_KEY environment variables or use cv2.set_server() before use."
            )
    return _global_cv2_server


def set_server(server):
    """Set the server to use for the cv2 frontend."""
    global _global_cv2_server
    assert isinstance(server, vf.Server)
    _global_cv2_server = server


def get_server():
    """Get the server used by the cv2 frontend."""
    return _server()


_PIX_FMT_MAP = {
    "rgb24": "rgb24",
    "yuv420p": "rgb24",
    "yuv422p": "rgb24",
    "yuv422p10le": "rgb24",
    "yuv444p": "rgb24",
    "yuvj420p": "rgb24",
    "yuvj422p": "rgb24",
    "yuvj444p": "rgb24",
    "gray": "gray",
}


def _top_level_pix_fmt(pix_fmt):
    if pix_fmt in _PIX_FMT_MAP:
        return _PIX_FMT_MAP[pix_fmt]
    raise Exception(f"Unsupported pix_fmt {pix_fmt}")


class Frame:
    def __init__(self, f, fmt, parent=None, slice_bounds=None):
        self._f_internal = f
        self._fmt = fmt
        channels = 3 if _top_level_pix_fmt(fmt["pix_fmt"]) == "rgb24" else 1
        self.shape = (fmt["height"], fmt["width"], channels)

        # denotes that the frame has not yet been modified
        # when a frame is modified, it is converted to rgb24 first
        self._modified = False

        # For slice views: track parent frame and bounds for write-back
        self._parent = parent
        self._slice_bounds = slice_bounds  # (miny, maxy, minx, maxx)

    @property
    def _f(self):
        return self._f_internal

    @_f.setter
    def _f(self, value):
        self._f_internal = value
        # If this is a slice view, propagate changes back to parent
        if self._parent is not None:
            miny, maxy, minx, maxx = self._slice_bounds
            self._parent._mut()
            self._parent._f = _slice_write_mat(
                self._parent._f, self._f_internal, miny, maxy, minx, maxx
            )

    def _mut(self):
        if self._modified:
            assert self._fmt["pix_fmt"] in ["rgb24", "gray"]
            return

        self._modified = True
        if (
            self._fmt["pix_fmt"] != "rgb24"
            and _top_level_pix_fmt(self._fmt["pix_fmt"]) == "rgb24"
        ):
            self._f = _filter_scale(self._f, pix_fmt="rgb24")
            self._fmt["pix_fmt"] = "rgb24"
        elif (
            self._fmt["pix_fmt"] != "gray"
            and _top_level_pix_fmt(self._fmt["pix_fmt"]) == "gray"
        ):
            self._f = _filter_scale(self._f, pix_fmt="gray")
            self._fmt["pix_fmt"] = "gray"

    def copy(self):
        # Copy creates an independent frame (no parent reference)
        return Frame(self._f, self._fmt.copy())

    def numpy(self):
        """
        Return the frame as a numpy array.
        """

        self._mut()
        server = _server()
        frame = server.frame(
            self.shape[1], self.shape[0], self._fmt["pix_fmt"], self._f
        )
        assert type(frame) is bytes
        assert len(frame) == self.shape[0] * self.shape[1] * self.shape[2]
        raw_data_array = np.frombuffer(frame, dtype=np.uint8)
        frame = raw_data_array.reshape(self.shape)
        if self.shape[2] == 3:
            frame = frame[:, :, ::-1]  # convert RGB to BGR
        return frame

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            raise NotImplementedError("Only 2D slicing is supported")

        if len(key) != 2:
            raise NotImplementedError("Only 2D slicing is supported")

        if not all(isinstance(x, slice) for x in key):
            raise NotImplementedError("Only 2D slicing is supported")

        miny = key[0].start if key[0].start is not None else 0
        maxy = key[0].stop if key[0].stop is not None else self.shape[0]
        minx = key[1].start if key[1].start is not None else 0
        maxx = key[1].stop if key[1].stop is not None else self.shape[1]

        # handle negative indices
        if miny < 0:
            miny = self.shape[0] + miny
        if maxy < 0:
            maxy = self.shape[0] + maxy
        if minx < 0:
            minx = self.shape[1] + minx
        if maxx < 0:
            maxx = self.shape[1] + maxx

        if (
            maxy <= miny
            or maxx <= minx
            or miny < 0
            or minx < 0
            or maxy > self.shape[0]
            or maxx > self.shape[1]
        ):
            raise NotImplementedError("Invalid slice")

        f = _slice_mat(self._f, miny, maxy, minx, maxx)
        fmt = self._fmt.copy()
        fmt["width"] = maxx - minx
        fmt["height"] = maxy - miny
        # Create slice with parent reference for write-back propagation
        return Frame(f, fmt, parent=self, slice_bounds=(miny, maxy, minx, maxx))

    def __setitem__(self, key, value):
        if type(key) is tuple:
            if len(key) != 2:
                raise NotImplementedError("Only 2D slicing is supported")

            if not all(isinstance(x, slice) for x in key):
                raise NotImplementedError("Only 2D slicing is supported")

            miny = key[0].start if key[0].start is not None else 0
            maxy = key[0].stop if key[0].stop is not None else self.shape[0]
            minx = key[1].start if key[1].start is not None else 0
            maxx = key[1].stop if key[1].stop is not None else self.shape[1]

            # handle negative indices
            if miny < 0:
                miny = self.shape[0] + miny
            if maxy < 0:
                maxy = self.shape[0] + maxy
            if minx < 0:
                minx = self.shape[1] + minx
            if maxx < 0:
                maxx = self.shape[1] + maxx

            if (
                maxy <= miny
                or maxx <= minx
                or miny < 0
                or minx < 0
                or maxy > self.shape[0]
                or maxx > self.shape[1]
            ):
                raise NotImplementedError("Invalid slice")

            # Check if value is a color tuple/list or scalar (solid fill)
            if isinstance(value, (int, float)):
                # Scalar value - treat as grayscale
                height = maxy - miny
                width = maxx - minx
                v = int(value)
                value = solid((height, width, self.shape[2]), (v, v, v))
            elif isinstance(value, (tuple, list)) and len(value) in [3, 4]:
                # Create a solid color frame of the right size
                height = maxy - miny
                width = maxx - minx
                value = solid((height, width, self.shape[2]), value[:3])
            else:
                value = frameify(value, "value")

            if value.shape[0] != maxy - miny or value.shape[1] != maxx - minx:
                raise NotImplementedError("Shape mismatch")

            self._mut()
            value._mut()

            self._f = _slice_write_mat(self._f, value._f, miny, maxy, minx, maxx)
        elif type(key) is Frame or type(key) is np.ndarray:
            key = frameify(key, "key")

            if key.shape[0] != self.shape[0] or key.shape[1] != self.shape[1]:
                raise NotImplementedError("Shape mismatch")

            if key.shape[2] != 1:
                raise NotImplementedError("Only 1-channel mask frames are supported")

            # Value should be a bgr or bgra color
            if (type(value) is not list and type(value) is not tuple) or len(
                value
            ) not in [3, 4]:
                raise NotImplementedError(
                    "Value should be a 3 or 4 element list or tuple"
                )
            value = [float(x) for x in value]
            if len(value) == 3:
                value.append(255.0)

            self._mut()
            key._mut()

            self._f = _set_to(self._f, value, key._f)
        else:
            raise NotImplementedError(
                "__setitem__ only supports slicing by a 2d tuple or a mask frame"
            )


def _inline_frame(arr):
    if arr.dtype != np.uint8:
        raise Exception("Only uint8 arrays are supported")
    if len(arr.shape) != 3:
        raise Exception("Only 3D arrays are supported")
    if arr.shape[2] != 3:
        raise Exception("To inline a frame, the array must have 3 channels")

    arr = arr[:, :, ::-1]
    if not arr.flags["C_CONTIGUOUS"]:
        arr = np.ascontiguousarray(arr)

    width = arr.shape[1]
    height = arr.shape[0]
    pix_fmt = "rgb24"

    data_gzip = zlib.compress(memoryview(arr), level=1)

    f = _inline_mat(
        data_gzip, width=width, height=height, pix_fmt=pix_fmt, compression="zlib"
    )
    fmt = {"width": width, "height": height, "pix_fmt": pix_fmt}

    # Return the resulting Frame object
    return Frame(f, fmt)


def _check_opencv2(method_name):
    if _opencv2 is None:
        raise NotImplementedError(
            f"{method_name} requires Python OpenCV cv2. Either it's not installed or the import failed (such as a missing libGL.so.1)."
        )


class VideoCapture:
    def __init__(self, path: str):
        server = _server()
        if type(path) is str:
            match = re.match(r"(http|https)://([^/]+)(.*)", path)
            if match is not None:
                endpoint = f"{match.group(1)}://{match.group(2)}"
                path = unquote(match.group(3))
                if path.startswith("/"):
                    path = path[1:]
                self._path = path
                self._source = server.source(path, 0, "http", {"endpoint": endpoint})
            else:
                self._path = path
                self._source = server.source(path, 0, "fs", {"root": "."})
        elif isinstance(path, vf.Source):
            assert isinstance(server, vf.Server)
            self._path = path._name
            self._source = path
        self._next_frame_idx = 0

    def isOpened(self) -> bool:
        return True

    def get(self, prop):
        if prop == CAP_PROP_FPS:
            return _ts_to_fps(self._source.ts())
        elif prop == CAP_PROP_FRAME_WIDTH:
            return self._source.fmt()["width"]
        elif prop == CAP_PROP_FRAME_HEIGHT:
            return self._source.fmt()["height"]
        elif prop == CAP_PROP_FRAME_COUNT:
            return len(self._source)
        elif prop == CAP_PROP_POS_FRAMES:
            return self._next_frame_idx
        elif prop == CAP_PROP_POS_MSEC:
            ts = self._source.ts()
            if self._next_frame_idx >= len(ts):
                # Past the end, return the last timestamp
                if len(ts) > 0:
                    return float(ts[-1] * 1000)
                return 0.0
            return float(ts[self._next_frame_idx] * 1000)

        raise Exception(f"Unknown property {prop}")

    def set(self, prop, value):
        if prop == CAP_PROP_POS_FRAMES:
            assert value >= 0 and value < len(self._source.ts())
            self._next_frame_idx = value
        elif prop == CAP_PROP_POS_MSEC:
            t = Fraction(int(value), 1000)
            ts = self._source.ts()
            next_frame_idx = bisect_right(ts, t)
            self._next_frame_idx = next_frame_idx
        else:
            raise Exception(f"Unsupported property {prop}")

    def read(self):
        if self._next_frame_idx >= len(self._source):
            return False, None
        frame = self._source.iloc[self._next_frame_idx]
        self._next_frame_idx += 1
        frame = Frame(frame, self._source.fmt())
        return True, frame

    def __getitem__(self, key):
        if not isinstance(key, int):
            raise NotImplementedError("Only integer indexing is supported")
        if key < 0:
            key = len(self._source) + key
        if key < 0 or key >= len(self._source):
            raise IndexError("Index out of bounds")
        frame = self._source.iloc[key]
        frame = Frame(frame, self._source.fmt())
        return frame

    def __len__(self):
        return len(self._source)

    def release(self):
        pass


class VideoWriter:
    def __init__(
        self,
        path,
        fourcc,
        fps,
        size,
        batch_size=1024,
        initial_batch_size=100,
        compression="gzip",
        ttl=3600,
        pix_fmt="yuv420p",
        vod_segment_length=Fraction(2, 1),
    ):
        server = _server()
        assert isinstance(server, vf.Server)
        assert path is None or type(path) is str
        if path is not None and server.is_vod_only():
            path = None

        self._path = path
        if isinstance(fps, int):
            self._f_time = Fraction(1, fps)
        elif isinstance(fps, Fraction):
            self._f_time = 1 / fps
        elif isinstance(fps, float):
            # 29.97
            if abs(fps - 30000 / 1001) < 1e-6:
                self._f_time = Fraction(1001, 30000)
            # 23.976
            elif abs(fps - 24000 / 1001) < 1e-6:
                self._f_time = Fraction(1001, 24000)
            # 59.94
            elif abs(fps - 60000 / 1001) < 1e-6:
                self._f_time = Fraction(1001, 60000)
            else:
                # Round to nearest integer fps
                self._f_time = Fraction(1, int(round(fps)))
        else:
            raise Exception("fps must be an integer, float, or Fraction")

        assert isinstance(size, tuple) or isinstance(size, list)
        assert len(size) == 2
        width, height = size
        assert ttl is None or isinstance(ttl, int)
        self._spec = server.create_spec(
            width, height, pix_fmt, vod_segment_length, 1 / self._f_time, ttl=ttl
        )
        self._batch_size = batch_size
        self._initial_batch_size = initial_batch_size
        assert compression is None or compression in ["gzip"]
        self._compression = compression
        self._idx = 0
        self._feb = vf._FrameExpressionBlock()
        self._first_push_done = False

        # writer_init_callback
        if server.cv2_writer_init_callback() is not None:
            server.cv2_writer_init_callback()(self)

    def _flush(self, terminal=False):
        server = _server()
        if len(self._feb) > 0:
            server.push_spec_part_block(
                self._spec,
                self._idx - len(self._feb),
                [self._feb],
                terminal=terminal,
                compression=self._compression,
            )
            self._feb = vf._FrameExpressionBlock()
        else:
            server.push_spec_part_block(
                self._spec,
                self._idx - len(self._feb),
                [],
                terminal=terminal,
            )

        # First push callback (called once after first successful push)
        if not self._first_push_done:
            self._first_push_done = True
            if server.cv2_writer_first_push_callback() is not None:
                server.cv2_writer_first_push_callback()(self)

    def spec(self):
        return self._spec

    def write(self, frame):
        if frame is not None:
            frame = frameify(frame, "frame")
            if frame._fmt["width"] != self._spec._fmt["width"]:
                raise Exception(
                    f"Frame type error; expected width {self._spec._fmt['width']}, got {frame._fmt['width']}"
                )
            if frame._fmt["height"] != self._spec._fmt["height"]:
                raise Exception(
                    f"Frame type error; expected height {self._spec._fmt['height']}, got {frame._fmt['height']}"
                )
            if frame._fmt["pix_fmt"] != self._spec._fmt["pix_fmt"]:
                f_obj = _filter_scale(frame._f, pix_fmt=self._spec._fmt["pix_fmt"])
                frame = Frame(f_obj, self._spec._fmt)
        self._feb.insert_frame(frame._f if frame is not None else None)
        self._idx += 1

        # Use smaller initial batch for faster first response
        threshold = (
            self._initial_batch_size if not self._first_push_done else self._batch_size
        )
        if len(self._feb) >= threshold:
            self._flush()

    def isOpened(self):
        return True

    def release(self):
        self._flush(True)
        if self._path is not None:
            server = _server()
            server.export_spec(self._spec.id(), self._path)


class VideoWriter_fourcc:
    def __init__(self, *args):
        self._args = args


def frameify(obj, field_name=None):
    """
    Turn an object (e.g., ndarray) into a Frame.
    """

    if isinstance(obj, Frame):
        return obj
    elif isinstance(obj, np.ndarray):
        return _inline_frame(obj)
    else:
        if field_name is not None:
            raise Exception(
                f"Unsupported type for field {field_name}, expected Frame or np.ndarray"
            )
        else:
            raise Exception("Unsupported type, expected Frame or np.ndarray")


def imread(path, *args):
    if len(args) > 0:
        raise NotImplementedError("imread does not support additional arguments")
    assert path.lower().endswith((".jpg", ".jpeg", ".png"))
    server = _server()

    cap = VideoCapture(path)
    assert cap.isOpened()
    assert len(cap._source) == 1
    ret, frame = cap.read()
    assert ret
    cap.release()
    return frame


def imwrite(path, img, *args):
    if len(args) > 0:
        raise NotImplementedError("imwrite does not support additional arguments")

    img = frameify(img)
    fmt = img._fmt.copy()
    width = fmt["width"]
    height = fmt["height"]

    if path.lower().endswith(".png"):
        out_pix_fmt = "rgb24"
        encoder = "png"
    elif path.lower().endswith((".jpg", ".jpeg")):
        encoder = "mjpeg"
        if img._fmt["pix_fmt"] not in ["yuvj420p", "yuvj422p", "yuvj444p"]:
            out_pix_fmt = "yuvj420p"
        else:
            out_pix_fmt = img._fmt["pix_fmt"]
    else:
        raise Exception("Unsupported image format")

    if img._fmt["pix_fmt"] != out_pix_fmt:
        f = _filter_scale(img._f, pix_fmt=out_pix_fmt)
        img = Frame(f, {"width": width, "height": height, "pix_fmt": out_pix_fmt})

    writer = VideoWriter(None, None, 1, (width, height), pix_fmt=out_pix_fmt)
    writer.write(img)
    writer.release()

    spec = writer.spec()
    server = _server()
    server.export_spec(spec.id(), path, encoder=encoder)


def vidplay(video, method="display"):
    """
    Play a vidformer video specification.
    """
    if isinstance(video, VideoWriter):
        return video.spec().play(method=method)
    elif isinstance(video, vf.Spec):
        return video.play(method=method)
    else:
        raise Exception("Unsupported video type to vidplay")


def zeros(shape, dtype=np.uint8):
    """
    Create a black frame. Mimics numpy.zeros.
    """
    assert isinstance(shape, tuple) or isinstance(shape, list)
    assert len(shape) == 3
    assert shape[2] in [1, 3]
    assert dtype == np.uint8

    height, width, channels = shape
    if channels == 1:
        pix_fmt = "gray"
    else:
        pix_fmt = "rgb24"

    f = _black(width=width, height=height, pix_fmt=pix_fmt)
    fmt = {"width": width, "height": height, "pix_fmt": pix_fmt}
    return Frame(f, fmt)


def solid(shape, color, dtype=np.uint8):
    """
    Create a solid color frame.

    Parameters:
        shape: (height, width, channels) tuple
        color: (B, G, R) tuple for BGR color (OpenCV convention)
        dtype: must be np.uint8
    """
    assert isinstance(shape, tuple) or isinstance(shape, list)
    assert len(shape) == 3
    assert shape[2] in [1, 3]
    assert dtype == np.uint8

    height, width, channels = shape
    if channels == 1:
        pix_fmt = "gray"
    else:
        pix_fmt = "rgb24"

    # Convert BGR to RGB for internal storage
    if len(color) == 3:
        r, g, b = color[2], color[1], color[0]
    else:
        r, g, b = color[0], color[1], color[2]

    f = _solid(width=width, height=height, pix_fmt=pix_fmt, color=[r, g, b])
    fmt = {"width": width, "height": height, "pix_fmt": pix_fmt}
    return Frame(f, fmt)


def resize(src, dsize, interpolation=None):
    src = frameify(src)
    src._mut()

    assert isinstance(dsize, tuple) or isinstance(dsize, list)
    assert len(dsize) == 2
    width, height = dsize

    # TODO: We don't do anything with interpolation yet
    assert interpolation is None or (
        interpolation >= INTER_NEAREST and interpolation <= INTER_MAX
    )

    f = _filter_scale(src._f, width=width, height=height)
    fmt = {"width": width, "height": height, "pix_fmt": src._fmt["pix_fmt"]}
    return Frame(f, fmt)


def rectangle(img, pt1, pt2, color, thickness=None, lineType=None, shift=None):
    """
    cv.rectangle(	img, pt1, pt2, color[, thickness[, lineType[, shift]]]	)
    """

    img = frameify(img)
    img._mut()

    assert len(pt1) == 2
    assert len(pt2) == 2
    pt1 = [int(x) for x in pt1]
    pt2 = [int(x) for x in pt2]

    assert len(color) == 3 or len(color) == 4
    color = [float(x) for x in color]
    if len(color) == 3:
        color.append(255.0)

    if shift is not None and lineType is None:
        lineType = LINE_8
    if lineType is not None and thickness is None:
        thickness = 1

    args = []
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if shift is not None:
        assert isinstance(shift, int)
        args.append(shift)

    img._f = _filter_rectangle(img._f, pt1, pt2, color, *args)
    return img


def putText(
    img,
    text,
    org,
    fontFace,
    fontScale,
    color,
    thickness=None,
    lineType=None,
    bottomLeftOrigin=None,
):
    """
    cv.putText(	img, text, org, fontFace, fontScale, color[, thickness[, lineType[, bottomLeftOrigin]]]	)
    """

    img = frameify(img)
    img._mut()

    assert isinstance(text, str)

    assert len(org) == 2
    org = [int(x) for x in org]

    assert isinstance(fontFace, int)
    assert isinstance(fontScale, float) or isinstance(fontScale, int)
    fontScale = float(fontScale)

    assert len(color) == 3 or len(color) == 4
    color = [float(x) for x in color]
    if len(color) == 3:
        color.append(255.0)

    if bottomLeftOrigin is not None and lineType is None:
        lineType = LINE_8
    if lineType is not None and thickness is None:
        thickness = 1

    args = []
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if bottomLeftOrigin is not None:
        assert isinstance(bottomLeftOrigin, bool)
        args.append(bottomLeftOrigin)

    img._f = _filter_putText(img._f, text, org, fontFace, fontScale, color, *args)
    return img


def arrowedLine(
    img, pt1, pt2, color, thickness=None, line_type=None, shift=None, tipLength=None
):
    """
    cv.arrowedLine(	img, pt1, pt2, color[, thickness[, line_type[, shift[, tipLength]]]]	)
    """
    img = frameify(img)
    img._mut()

    assert len(pt1) == 2
    assert len(pt2) == 2
    assert all(isinstance(x, int) for x in pt1)
    assert all(isinstance(x, int) for x in pt2)

    assert len(color) == 3 or len(color) == 4
    color = [float(x) for x in color]
    if len(color) == 3:
        color.append(255.0)

    if tipLength is not None and shift is None:
        shift = 0
    if shift is not None and line_type is None:
        line_type = LINE_8
    if line_type is not None and thickness is None:
        thickness = 1

    args = []
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if line_type is not None:
        assert isinstance(line_type, int)
        args.append(line_type)
    if shift is not None:
        assert isinstance(shift, int)
        args.append(shift)
    if tipLength is not None:
        assert isinstance(tipLength, (int, float))
        args.append(float(tipLength))

    img._f = _filter_arrowedLine(img._f, pt1, pt2, color, *args)
    return img


def line(img, pt1, pt2, color, thickness=None, lineType=None, shift=None):
    img = frameify(img)
    img._mut()

    assert len(pt1) == 2
    assert len(pt2) == 2
    pt1 = [int(x) for x in pt1]
    pt2 = [int(x) for x in pt2]

    assert len(color) == 3 or len(color) == 4
    color = [float(x) for x in color]
    if len(color) == 3:
        color.append(255.0)

    if shift is not None and lineType is None:
        lineType = LINE_8
    if lineType is not None and thickness is None:
        thickness = 1

    args = []
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if shift is not None:
        assert isinstance(shift, int)
        args.append(shift)

    img._f = _filter_line(img._f, pt1, pt2, color, *args)
    return img


def circle(img, center, radius, color, thickness=None, lineType=None, shift=None):
    img = frameify(img)
    img._mut()

    assert len(center) == 2
    center = [int(x) for x in center]

    assert isinstance(radius, int)

    assert len(color) == 3 or len(color) == 4
    color = [float(x) for x in color]
    if len(color) == 3:
        color.append(255.0)

    if shift is not None and lineType is None:
        lineType = LINE_8
    if lineType is not None and thickness is None:
        thickness = 1

    args = []
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if shift is not None:
        assert isinstance(shift, int)
        args.append(shift)

    img._f = _filter_circle(img._f, center, radius, color, *args)
    return img


def getFontScaleFromHeight(*args, **kwargs):
    """
    cv.getFontScaleFromHeight(	fontFace, pixelHeight[, thickness]	)
    """
    _check_opencv2("getFontScaleFromHeight")
    return _opencv2.getFontScaleFromHeight(*args, **kwargs)


def getTextSize(*args, **kwargs):
    """
    cv.getTextSize(	text, fontFace, fontScale, thickness	)
    """
    _check_opencv2("getTextSize")
    return _opencv2.getTextSize(*args, **kwargs)


def addWeighted(src1, alpha, src2, beta, gamma, dst=None, dtype=-1):
    """
    cv.addWeighted(	src1, alpha, src2, beta, gamma[, dst[, dtype]]	) -> 	dst
    """
    src1 = frameify(src1, "src1")
    src2 = frameify(src2, "src2")
    src1._mut()
    src2._mut()

    if dst is None:
        dst = Frame(src1._f, src1._fmt.copy())
    else:
        assert isinstance(dst, Frame), "dst must be a Frame"
    dst._mut()

    assert isinstance(alpha, float) or isinstance(alpha, int)
    assert isinstance(beta, float) or isinstance(beta, int)
    assert isinstance(gamma, float) or isinstance(gamma, int)
    alpha = float(alpha)
    beta = float(beta)
    gamma = float(gamma)

    if dtype != -1:
        raise Exception("addWeighted does not support the dtype argument")

    dst._f = _filter_addWeighted(src1._f, alpha, src2._f, beta, gamma)
    return dst


def ellipse(
    img,
    center,
    axes,
    angle,
    startAngle,
    endAngle,
    color,
    thickness=1,
    lineType=LINE_8,
    shift=0,
):
    img = frameify(img)
    img._mut()

    assert len(center) == 2
    center = [int(x) for x in center]

    assert len(axes) == 2
    axes = [int(x) for x in axes]

    assert isinstance(angle, float) or isinstance(angle, int)
    assert isinstance(startAngle, float) or isinstance(startAngle, int)
    assert isinstance(endAngle, float) or isinstance(endAngle, int)
    angle = float(angle)
    startAngle = float(startAngle)
    endAngle = float(endAngle)

    assert len(color) == 3 or len(color) == 4
    color = [float(x) for x in color]
    if len(color) == 3:
        color.append(255.0)

    assert isinstance(thickness, int)
    assert isinstance(lineType, int)
    assert isinstance(shift, int)

    img._f = _filter_ellipse(
        img._f,
        center,
        axes,
        angle,
        startAngle,
        endAngle,
        color,
        thickness,
        lineType,
        shift,
    )
    return img


# Stubs for unimplemented functions


def clipLine(*args, **kwargs):
    """
    cv.clipLine(imgRect, pt1, pt2) -> retval, pt1, pt2
    """
    _check_opencv2("clipLine")
    return _opencv2.clipLine(*args, **kwargs)


def drawContours(
    img,
    contours,
    contourIdx,
    color,
    thickness=None,
    lineType=None,
    hierarchy=None,
    maxLevel=None,
    offset=None,
):
    """
    cv.drawContours(image, contours, contourIdx, color[, thickness[, lineType[, hierarchy[, maxLevel[, offset]]]]]) -> image
    """
    img = frameify(img)
    img._mut()

    # Convert contours to list format
    contours_converted = []
    for contour in contours:
        contours_converted.append(_convert_single_polygon(contour))

    assert isinstance(contourIdx, int)
    color = _convert_color(color)

    if offset is not None and maxLevel is None:
        maxLevel = 0
    if maxLevel is not None and hierarchy is None:
        hierarchy = []
    if hierarchy is not None and lineType is None:
        lineType = LINE_8
    if lineType is not None and thickness is None:
        thickness = 1

    args = []
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if hierarchy is not None:
        if isinstance(hierarchy, np.ndarray):
            hierarchy = hierarchy.tolist()
        args.append(hierarchy)
    if maxLevel is not None:
        assert isinstance(maxLevel, int)
        args.append(maxLevel)
    if offset is not None:
        assert isinstance(offset, (list, tuple)) and len(offset) == 2
        args.append([int(offset[0]), int(offset[1])])

    img._f = _filter_drawContours(img._f, contours_converted, contourIdx, color, *args)
    return img


def drawMarker(
    img,
    position,
    color,
    markerType=None,
    markerSize=None,
    thickness=None,
    line_type=None,
):
    """
    cv.drawMarker(img, position, color[, markerType[, markerSize[, thickness[, line_type]]]]) -> img
    """
    img = frameify(img)
    img._mut()

    assert isinstance(position, (list, tuple)) and len(position) == 2
    position = [int(position[0]), int(position[1])]
    color = _convert_color(color)

    if line_type is not None and thickness is None:
        thickness = 1
    if thickness is not None and markerSize is None:
        markerSize = 20
    if markerSize is not None and markerType is None:
        markerType = MARKER_CROSS

    args = []
    if markerType is not None:
        assert isinstance(markerType, int)
        args.append(markerType)
    if markerSize is not None:
        assert isinstance(markerSize, int)
        args.append(markerSize)
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if line_type is not None:
        assert isinstance(line_type, int)
        args.append(line_type)

    img._f = _filter_drawMarker(img._f, position, color, *args)
    return img


def ellipse2Poly(*args, **kwargs):
    """
    cv.ellipse2Poly(center, axes, angle, arcStart, arcEnd, delta) -> pts
    """
    _check_opencv2("ellipse2Poly")
    return _opencv2.ellipse2Poly(*args, **kwargs)


def _convert_polygon_list(pts):
    """Convert a list of polygons to the internal format."""
    assert isinstance(pts, list) or isinstance(pts, np.ndarray)
    pts_converted = []
    for poly in pts:
        pts_converted.append(_convert_single_polygon(poly))
    return pts_converted


def _convert_single_polygon(poly):
    """Convert a single polygon to the internal format."""
    if isinstance(poly, np.ndarray):
        poly = poly.tolist()
    # Flatten if shape is (N, 1, 2)
    poly_flat = []
    for pt in poly:
        if isinstance(pt, list) and len(pt) == 1:
            pt = pt[0]
        poly_flat.append([int(pt[0]), int(pt[1])])
    return poly_flat


def _convert_color(color):
    """Convert color to the internal format (4 floats)."""
    assert len(color) == 3 or len(color) == 4
    color = [float(x) for x in color]
    if len(color) == 3:
        color.append(255.0)
    return color


def fillConvexPoly(img, points, color, lineType=None, shift=None):
    """
    cv.fillConvexPoly(img, points, color[, lineType[, shift]]) -> img
    """
    img = frameify(img)
    img._mut()

    points_converted = _convert_single_polygon(points)
    color = _convert_color(color)

    if shift is not None and lineType is None:
        lineType = LINE_8

    args = []
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if shift is not None:
        assert isinstance(shift, int)
        args.append(shift)

    img._f = _filter_fillConvexPoly(img._f, points_converted, color, *args)
    return img


def fillPoly(img, pts, color, lineType=None, shift=None, offset=None):
    """
    cv.fillPoly(img, pts, color[, lineType[, shift[, offset]]]) -> img
    """
    img = frameify(img)
    img._mut()

    pts_converted = _convert_polygon_list(pts)
    color = _convert_color(color)

    if offset is not None and shift is None:
        shift = 0
    if shift is not None and lineType is None:
        lineType = LINE_8

    args = []
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if shift is not None:
        assert isinstance(shift, int)
        args.append(shift)
    if offset is not None:
        assert isinstance(offset, (list, tuple)) and len(offset) == 2
        args.append([int(offset[0]), int(offset[1])])

    img._f = _filter_fillPoly(img._f, pts_converted, color, *args)
    return img


def polylines(img, pts, isClosed, color, thickness=None, lineType=None, shift=None):
    """
    cv.polylines(img, pts, isClosed, color[, thickness[, lineType[, shift]]]) -> img
    """
    img = frameify(img)
    img._mut()

    pts_converted = _convert_polygon_list(pts)
    assert isinstance(isClosed, bool)
    color = _convert_color(color)

    if shift is not None and lineType is None:
        lineType = LINE_8
    if lineType is not None and thickness is None:
        thickness = 1

    args = []
    if thickness is not None:
        assert isinstance(thickness, int)
        args.append(thickness)
    if lineType is not None:
        assert isinstance(lineType, int)
        args.append(lineType)
    if shift is not None:
        assert isinstance(shift, int)
        args.append(shift)

    img._f = _filter_polylines(img._f, pts_converted, isClosed, color, *args)
    return img


def flip(src, flipCode):
    """
    cv.flip(src, flipCode[, dst]) -> dst

    Flips a 2D array around vertical, horizontal, or both axes.

    Parameters:
        src: input array
        flipCode: a flag to specify how to flip the array:
            0 means flipping around the x-axis (vertical flip)
            positive value (e.g., 1) means flipping around y-axis (horizontal flip)
            negative value (e.g., -1) means flipping around both axes
    """
    src = frameify(src)
    src._mut()

    assert isinstance(flipCode, int)

    f = _filter_flip(src._f, flipCode)
    return Frame(f, src._fmt.copy())


def rotate(src, rotateCode):
    """
    cv.rotate(src, rotateCode[, dst]) -> dst

    Rotates a 2D array in multiples of 90 degrees.

    Parameters:
        src: input array
        rotateCode: an enum to specify how to rotate the array:
            cv2.ROTATE_90_CLOCKWISE (0) - rotate 90 degrees clockwise
            cv2.ROTATE_180 (1) - rotate 180 degrees
            cv2.ROTATE_90_COUNTERCLOCKWISE (2) - rotate 90 degrees counter-clockwise
    """
    src = frameify(src)
    src._mut()

    assert isinstance(rotateCode, int)
    assert rotateCode in [ROTATE_90_CLOCKWISE, ROTATE_180, ROTATE_90_COUNTERCLOCKWISE]

    # Calculate new dimensions
    width = src._fmt["width"]
    height = src._fmt["height"]

    if rotateCode == ROTATE_180:
        new_width, new_height = width, height
    else:
        # 90 degree rotations swap width and height
        new_width, new_height = height, width

    f = _filter_rotate(src._f, rotateCode)
    fmt = {"width": new_width, "height": new_height, "pix_fmt": src._fmt["pix_fmt"]}
    return Frame(f, fmt)


def copyMakeBorder(src, top, bottom, left, right, borderType, value=None):
    """
    cv.copyMakeBorder(src, top, bottom, left, right, borderType[, dst[, value]]) -> dst

    Forms a border around an image.

    Parameters:
        src: input array
        top: number of pixels for the top border
        bottom: number of pixels for the bottom border
        left: number of pixels for the left border
        right: number of pixels for the right border
        borderType: border type (cv2.BORDER_CONSTANT, cv2.BORDER_REPLICATE, etc.)
        value: border value if borderType == cv2.BORDER_CONSTANT (default: black)
    """
    src = frameify(src)
    src._mut()

    assert isinstance(top, int) and top >= 0
    assert isinstance(bottom, int) and bottom >= 0
    assert isinstance(left, int) and left >= 0
    assert isinstance(right, int) and right >= 0
    assert isinstance(borderType, int)

    # Handle value parameter
    if value is None:
        value = [0.0, 0.0, 0.0, 255.0]
    else:
        assert len(value) == 3 or len(value) == 4
        value = [float(x) for x in value]
        if len(value) == 3:
            value.append(255.0)

    # Calculate new dimensions
    new_width = src._fmt["width"] + left + right
    new_height = src._fmt["height"] + top + bottom

    f = _filter_copyMakeBorder(src._f, top, bottom, left, right, borderType, value)
    fmt = {"width": new_width, "height": new_height, "pix_fmt": src._fmt["pix_fmt"]}
    return Frame(f, fmt)


def hconcat(src):
    """
    cv.hconcat(src[, dst]) -> dst

    Concatenates arrays horizontally.

    Parameters:
        src: sequence of arrays to concatenate. All arrays must have the same height.
    """
    assert isinstance(src, (list, tuple)) and len(src) > 0

    frames = []
    height = None
    pix_fmt = None

    for item in src:
        item = frameify(item)
        item._mut()

        if height is None:
            height = item._fmt["height"]
            pix_fmt = item._fmt["pix_fmt"]
        else:
            assert (
                item._fmt["height"] == height
            ), "All frames must have the same height for hconcat"

        frames.append(item._f)

    # Calculate total width
    total_width = sum(frameify(item)._fmt["width"] for item in src)

    f = _filter_hconcat([frame for frame in frames])
    fmt = {"width": total_width, "height": height, "pix_fmt": pix_fmt}
    return Frame(f, fmt)


def vconcat(src):
    """
    cv.vconcat(src[, dst]) -> dst

    Concatenates arrays vertically.

    Parameters:
        src: sequence of arrays to concatenate. All arrays must have the same width.
    """
    assert isinstance(src, (list, tuple)) and len(src) > 0

    frames = []
    width = None
    pix_fmt = None

    for item in src:
        item = frameify(item)
        item._mut()

        if width is None:
            width = item._fmt["width"]
            pix_fmt = item._fmt["pix_fmt"]
        else:
            assert (
                item._fmt["width"] == width
            ), "All frames must have the same width for vconcat"

        frames.append(item._f)

    # Calculate total height
    total_height = sum(frameify(item)._fmt["height"] for item in src)

    f = _filter_vconcat([frame for frame in frames])
    fmt = {"width": width, "height": total_height, "pix_fmt": pix_fmt}
    return Frame(f, fmt)



"""
vidformer-py is a Python 🐍 interface for [vidformer](https://github.com/ixlab/vidformer).

**Quick links:**
* [📦 PyPI](https://pypi.org/project/vidformer/)
* [📘 Documentation - vidformer-py](https://ixlab.github.io/vidformer/vidformer-py/)
* [📘 Documentation - vidformer.cv2](https://ixlab.github.io/vidformer/vidformer-py/vidformer/cv2.html)
* [📘 Documentation - vidformer.supervision](https://ixlab.github.io/vidformer/vidformer-py/vidformer/supervision.html)
* [🧑‍💻 Source Code](https://github.com/ixlab/vidformer/tree/main/vidformer-py/)
"""

__version__ = "1.4.2"


import base64
import gzip
import json
import struct
import time
from fractions import Fraction
from urllib.parse import urlparse

import requests

_in_notebook = False
try:
    from IPython import get_ipython

    if "IPKernelApp" in get_ipython().config:
        _in_notebook = True
except Exception:
    pass


def _wait_for_url(url, max_attempts=150, delay=0.1):
    for attempt in range(max_attempts):
        try:
            response = requests.get(url)
            if response.status_code == 200:
                return response.text.strip()
            else:
                time.sleep(delay)
        except requests.exceptions.RequestException:
            time.sleep(delay)
    return None


def _play(namespace, hls_video_url, hls_js_url, method="display", status_url=None):
    # The namespace is so multiple videos in one tab don't conflict

    if method == "display":
        from IPython.display import display

        display(
            _play(
                namespace,
                hls_video_url,
                hls_js_url,
                method="html",
                status_url=status_url,
            )
        )
        return
    if method == "html":
        from IPython.display import HTML

        if not status_url:
            html_code = f"""
<!DOCTYPE html>
<html>
<head>
    <title>HLS Video Player</title>
    <!-- Include hls.js library -->
    <script src="{hls_js_url}"></script>
</head>
<body>
    <video id="video-{namespace}" controls width="640" height="360" autoplay></video>
    <script>
        var video = document.getElementById('video-{namespace}');
        var videoSrc = '{hls_video_url}';

        if (Hls.isSupported()) {{
            var hls = new Hls();
            hls.loadSource(videoSrc);
            hls.attachMedia(video);
            hls.on(Hls.Events.MANIFEST_PARSED, function() {{
                video.play();
            }});
        }} else if (video.canPlayType('application/vnd.apple.mpegurl')) {{
            video.src = videoSrc;
            video.addEventListener('loadedmetadata', function() {{
                video.play();
            }});
        }} else {{
            console.error('This browser does not appear to support HLS.');
        }}
    </script>
</body>
</html>
"""
            return HTML(data=html_code)
        else:
            html_code = f"""
<!DOCTYPE html>
<html>
<head>
    <title>HLS Video Player</title>
    <script src="{hls_js_url}"></script>
</head>
<body>
    <div id="container-{namespace}"></div>
    <script>
        var statusUrl = '{status_url}';
        var videoSrc = '{hls_video_url}';
        var videoNamespace = '{namespace}';

        function showWaiting() {{
            document.getElementById('container-{namespace}').textContent = 'Waiting...';
            pollStatus();
        }}

        function pollStatus() {{
            setTimeout(function() {{
                fetch(statusUrl)
                    .then(r => r.json())
                    .then(res => {{
                        if (res.ready) {{
                            document.getElementById('container-{namespace}').textContent = '';
                            attachHls();
                        }} else {{
                            pollStatus();
                        }}
                    }})
                    .catch(e => {{
                        console.error(e);
                        pollStatus();
                    }});
            }}, 250);
        }}

        function attachHls() {{
            var container = document.getElementById('container-{namespace}');
            container.textContent = '';
            var video = document.createElement('video');
            video.id = 'video-' + videoNamespace;
            video.controls = true;
            video.width = 640;
            video.height = 360;
            container.appendChild(video);
            if (Hls.isSupported()) {{
                var hls = new Hls();
                hls.loadSource(videoSrc);
                hls.attachMedia(video);
                hls.on(Hls.Events.MANIFEST_PARSED, function() {{
                    video.play();
                }});
            }} else if (video.canPlayType('application/vnd.apple.mpegurl')) {{
                video.src = videoSrc;
                video.addEventListener('loadedmetadata', function() {{
                    video.play();
                }});
            }}
        }}

        fetch(statusUrl)
            .then(r => r.json())
            .then(res => {{
                if (res.ready) {{
                    attachHls();
                }} else {{
                    showWaiting();
                }}
            }})
            .catch(e => {{
                console.error(e);
                showWaiting();
            }});
    </script>
</body>
</html>
"""
        return HTML(data=html_code)
    elif method == "link":
        return hls_video_url
    else:
        raise ValueError("Invalid method")


def _feb_expr_coded_as_scalar(expr) -> bool:
    if type(expr) is tuple:
        expr = list(expr)
    if type(expr) is FilterExpr:
        return False
    if type(expr) is list:
        if len(expr) > 3:
            return False
        else:
            return all([type(x) is int and x >= -(2**15) and x < 2**15 for x in expr])
    else:
        assert type(expr) in [int, float, str, bytes, SourceExpr, bool, list]
        return True


class _FrameExpressionBlock:
    def __init__(self):
        self._functions = []
        self._literals = []
        self._sources = []
        self._kwarg_keys = []
        self._source_fracs = []
        self._exprs = []
        self._frame_exprs = []

    def __len__(self):
        return len(self._frame_exprs)

    def insert_expr(self, expr):
        if type(expr) is SourceExpr or type(expr) is FilterExpr:
            return self.insert_frame_expr(expr)
        else:
            return self.insert_data_expr(expr)

    def insert_data_expr(self, data):
        if type(data) is tuple:
            data = list(data)

        if type(data) is bool:
            self._exprs.append(0x01000000_00000000 | int(data))
            return len(self._exprs) - 1
        elif type(data) is int:
            if data >= -(2**31) and data < 2**31:
                self._exprs.append(data & 0xFFFFFFFF)
            else:
                self._literals.append(_json_arg(data, skip_data_anot=True))
                self._exprs.append(0x40000000_00000000 | len(self._literals) - 1)
            return len(self._exprs) - 1
        elif type(data) is float:
            self._exprs.append(
                0x02000000_00000000 | int.from_bytes(struct.pack("f", data)[::-1])
            )
        elif type(data) is str:
            self._literals.append(_json_arg(data, skip_data_anot=True))
            self._exprs.append(0x40000000_00000000 | len(self._literals) - 1)
        elif type(data) is bytes:
            self._literals.append(_json_arg(data, skip_data_anot=True))
            self._exprs.append(0x40000000_00000000 | len(self._literals) - 1)
        elif type(data) is list:
            if len(data) == 0:
                self._exprs.append(0x03000000_00000000)
                return len(self._exprs) - 1
            if (
                len(data) == 1
                and type(data[0]) is int
                and data[0] >= -(2**15)
                and data[0] < 2**15
            ):
                self._exprs.append(0x04000000_00000000 | (data[0] & 0xFFFF))
                return len(self._exprs) - 1
            if (
                len(data) == 2
                and type(data[0]) is int
                and data[0] >= -(2**15)
                and data[0] < 2**15
                and type(data[1]) is int
                and data[1] >= -(2**15)
                and data[1] < 2**15
            ):
                self._exprs.append(
                    0x05000000_00000000
                    | ((data[0] & 0xFFFF) << 16)
                    | (data[1] & 0xFFFF)
                )
                return len(self._exprs) - 1
            if (
                len(data) == 3
                and type(data[0]) is int
                and data[0] >= -(2**15)
                and data[0] < 2**15
                and type(data[1]) is int
                and data[1] >= -(2**15)
                and data[1] < 2**15
                and type(data[2]) is int
                and data[2] >= -(2**15)
                and data[2] < 2**15
            ):
                self._exprs.append(
                    0x06000000_00000000
                    | ((data[0] & 0xFFFF) << 32)
                    | ((data[1] & 0xFFFF) << 16)
                    | (data[2] & 0xFFFF)
                )
                return len(self._exprs) - 1
            member_idxs = []
            for member in data:
                if _feb_expr_coded_as_scalar(member):
                    member_idxs.append(None)
                else:
                    member_idxs.append(self.insert_expr(member))

            out = len(self._exprs)
            self._exprs.append(0x42000000_00000000 | len(data))

            for i in range(len(data)):
                if member_idxs[i] is None:
                    self.insert_data_expr(data[i])
                else:
                    self._exprs.append(0x45000000_00000000 | member_idxs[i])

            return out
        else:
            raise Exception("Invalid data type")

    def insert_frame_expr(self, frame):
        if type(frame) is SourceExpr:
            source = frame._source._name
            if source in self._sources:
                source_idx = self._sources.index(source)
            else:
                source_idx = len(self._sources)
                self._sources.append(source)
            if frame._is_iloc:
                self._exprs.append(
                    0x43000000_00000000 | (source_idx << 32) | frame._idx
                )
            else:
                idx = len(self._source_fracs) // 2
                self._source_fracs.append(frame._idx.numerator)
                self._source_fracs.append(frame._idx.denominator)
                self._exprs.append(0x44000000_00000000 | (source_idx << 32) | idx)
            return len(self._exprs) - 1
        elif type(frame) is FilterExpr:
            func = frame._filter._func
            if func in self._functions:
                func_idx = self._functions.index(func)
            else:
                func_idx = len(self._functions)
                self._functions.append(func)
            len_args = len(frame._args)
            len_kwargs = len(frame._kwargs)

            arg_idxs = []
            for arg in frame._args:
                if _feb_expr_coded_as_scalar(arg):
                    arg_idxs.append(None)
                else:
                    arg_idxs.append(self.insert_expr(arg))
            kwarg_idxs = {}
            for k, v in frame._kwargs.items():
                if _feb_expr_coded_as_scalar(v):
                    kwarg_idxs[k] = None
                else:
                    kwarg_idxs[k] = self.insert_expr(v)

            out_idx = len(self._exprs)
            self._exprs.append(
                0x41000000_00000000 | (len_args << 24) | (len_kwargs << 16) | func_idx
            )
            for i in range(len_args):
                if arg_idxs[i] is None:
                    # It's a scalar
                    self.insert_expr(frame._args[i])
                else:
                    # It's an expression pointer
                    self._exprs.append(0x45000000_00000000 | arg_idxs[i])
            for k, v in frame._kwargs.items():
                if k in self._kwarg_keys:
                    k_idx = self._kwarg_keys.index(k)
                else:
                    k_idx = len(self._kwarg_keys)
                    self._kwarg_keys.append(k)
                self._exprs.append(0x46000000_00000000 | k_idx)
                if kwarg_idxs[k] is None:
                    # It's a scalar
                    self.insert_expr(v)
                else:
                    # It's an expression pointer
                    self._exprs.append(0x45000000_00000000 | kwarg_idxs[k])
            return out_idx
        else:
            raise Exception("Invalid frame type")

    def insert_frame(self, frame):
        idx = self.insert_frame_expr(frame)
        self._frame_exprs.append(idx)

    def as_dict(self):
        return {
            "functions": self._functions,
            "literals": self._literals,
            "sources": self._sources,
            "kwarg_keys": self._kwarg_keys,
            "source_fracs": self._source_fracs,
            "exprs": self._exprs,
            "frame_exprs": self._frame_exprs,
        }


class Source:
    def __init__(self, id: str, src):
        self._name = id
        self._fmt = {
            "width": src["width"],
            "height": src["height"],
            "pix_fmt": src["pix_fmt"],
        }
        self._ts = [Fraction(x[0], x[1]) for x in src["ts"]]
        self.iloc = _SourceILoc(self)

    def id(self) -> str:
        return self._name

    def fmt(self):
        return {**self._fmt}

    def ts(self) -> list[Fraction]:
        return self._ts.copy()

    def __len__(self):
        return len(self._ts)

    def __getitem__(self, idx):
        if type(idx) is not Fraction:
            raise Exception("Source index must be a Fraction")
        return SourceExpr(self, idx, False)

    def __repr__(self):
        return f"Source({self._name})"


class Spec:
    def __init__(self, id: str, src):
        self._id = id
        self._fmt = {
            "width": src["width"],
            "height": src["height"],
            "pix_fmt": src["pix_fmt"],
        }
        self._vod_endpoint = src["vod_endpoint"]
        parsed_url = urlparse(self._vod_endpoint)
        self._hls_js_url = f"{parsed_url.scheme}://{parsed_url.netloc}/hls.js"

    def id(self) -> str:
        return self._id

    def play(self, method):
        url = f"{self._vod_endpoint}playlist.m3u8"
        status_url = f"{self._vod_endpoint}status"
        hls_js_url = self._hls_js_url
        return _play(self._id, url, hls_js_url, method=method, status_url=status_url)


class Server:
    def __init__(
        self,
        endpoint: str,
        api_key: str,
        vod_only=False,
        cv2_writer_init_callback=None,
        cv2_writer_first_push_callback=None,
    ):
        if not endpoint.startswith("http://") and not endpoint.startswith("https://"):
            raise Exception("Endpoint must start with http:// or https://")
        if endpoint.endswith("/"):
            raise Exception("Endpoint must not end with /")
        self._endpoint = endpoint

        self._api_key = api_key
        self._session = requests.Session()
        self._session.headers.update({"Authorization": f"Bearer {self._api_key}"})
        response = self._session.get(
            f"{self._endpoint}/v2/auth",
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"
        self._vod_only = vod_only
        self._cv2_writer_init_callback = cv2_writer_init_callback
        self._cv2_writer_first_push_callback = cv2_writer_first_push_callback

    def is_vod_only(self) -> bool:
        return self._vod_only

    def cv2_writer_init_callback(self):
        return self._cv2_writer_init_callback

    def cv2_writer_first_push_callback(self):
        return self._cv2_writer_first_push_callback

    def get_source(self, id: str) -> Source:
        assert type(id) is str
        response = self._session.get(
            f"{self._endpoint}/v2/source/{id}",
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        return Source(response["id"], response)

    def list_sources(self) -> list[str]:
        response = self._session.get(
            f"{self._endpoint}/v2/source",
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        return response

    def delete_source(self, id: str):
        assert type(id) is str
        response = self._session.delete(
            f"{self._endpoint}/v2/source/{id}",
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"

    def search_source(
        self, name, stream_idx, storage_service, storage_config
    ) -> list[str]:
        assert type(name) is str
        assert type(stream_idx) is int
        assert type(storage_service) is str
        assert type(storage_config) is dict
        for k, v in storage_config.items():
            assert type(k) is str
            assert type(v) is str
        req = {
            "name": name,
            "stream_idx": stream_idx,
            "storage_service": storage_service,
            "storage_config": storage_config,
        }
        response = self._session.post(
            f"{self._endpoint}/v2/source/search",
            json=req,
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        return response

    def create_source(
        self, name, stream_idx, storage_service, storage_config
    ) -> Source:
        assert type(name) is str
        assert type(stream_idx) is int
        assert type(storage_service) is str
        assert type(storage_config) is dict
        for k, v in storage_config.items():
            assert type(k) is str
            assert type(v) is str
        req = {
            "name": name,
            "stream_idx": stream_idx,
            "storage_service": storage_service,
            "storage_config": storage_config,
        }
        response = self._session.post(
            f"{self._endpoint}/v2/source",
            json=req,
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"
        id = response["id"]
        return self.get_source(id)

    def source(self, name, stream_idx, storage_service, storage_config) -> Source:
        """Convenience function for accessing sources.

        Tries to find a source with the given name, stream_idx, storage_service, and storage_config.
        If no source is found, creates a new source with the given parameters.
        """

        sources = self.search_source(name, stream_idx, storage_service, storage_config)
        if len(sources) == 0:
            return self.create_source(name, stream_idx, storage_service, storage_config)
        return self.get_source(sources[0])

    def get_spec(self, id: str) -> Spec:
        assert type(id) is str
        response = self._session.get(
            f"{self._endpoint}/v2/spec/{id}",
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        return Spec(response["id"], response)

    def list_specs(self) -> list[str]:
        response = self._session.get(
            f"{self._endpoint}/v2/spec",
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        return response

    def create_spec(
        self,
        width,
        height,
        pix_fmt,
        vod_segment_length,
        frame_rate,
        ready_hook=None,
        steer_hook=None,
        ttl=None,
    ) -> Spec:
        assert type(width) is int
        assert type(height) is int
        assert type(pix_fmt) is str
        assert type(vod_segment_length) is Fraction
        assert type(frame_rate) is Fraction
        assert type(ready_hook) is str or ready_hook is None
        assert type(steer_hook) is str or steer_hook is None
        assert ttl is None or type(ttl) is int

        req = {
            "width": width,
            "height": height,
            "pix_fmt": pix_fmt,
            "vod_segment_length": [
                vod_segment_length.numerator,
                vod_segment_length.denominator,
            ],
            "frame_rate": [frame_rate.numerator, frame_rate.denominator],
            "ready_hook": ready_hook,
            "steer_hook": steer_hook,
            "ttl": ttl,
        }
        response = self._session.post(
            f"{self._endpoint}/v2/spec",
            json=req,
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"
        return self.get_spec(response["id"])

    def delete_spec(self, id: str):
        assert type(id) is str
        response = self._session.delete(
            f"{self._endpoint}/v2/spec/{id}",
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"

    def export_spec(
        self, id: str, path: str, encoder=None, encoder_opts=None, format=None
    ):
        assert type(id) is str
        assert type(path) is str
        req = {
            "path": path,
            "encoder": encoder,
            "encoder_opts": encoder_opts,
            "format": format,
        }
        response = self._session.post(
            f"{self._endpoint}/v2/spec/{id}/export",
            json=req,
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"

    def push_spec_part(self, spec_id, pos, frames, terminal):
        if type(spec_id) is Spec:
            spec_id = spec_id._id
        assert type(spec_id) is str
        assert type(pos) is int
        assert type(frames) is list
        assert type(terminal) is bool

        req_frames = []
        for frame in frames:
            assert type(frame) is tuple
            assert len(frame) == 2
            t = frame[0]
            f = frame[1]
            assert type(t) is Fraction
            assert f is None or type(f) is SourceExpr or type(f) is FilterExpr
            req_frames.append(
                [
                    [t.numerator, t.denominator],
                    f._to_json_spec() if f is not None else None,
                ]
            )

        req = {
            "pos": pos,
            "frames": req_frames,
            "terminal": terminal,
        }
        response = self._session.post(
            f"{self._endpoint}/v2/spec/{spec_id}/part",
            json=req,
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"

    def push_spec_part_block(
        self, spec_id: str, pos, blocks, terminal, compression="gzip"
    ):
        if type(spec_id) is Spec:
            spec_id = spec_id._id
        assert type(spec_id) is str
        assert type(pos) is int
        assert type(blocks) is list
        assert type(terminal) is bool
        assert compression is None or compression == "gzip"

        req_blocks = []
        for block in blocks:
            assert type(block) is _FrameExpressionBlock
            block_body = block.as_dict()
            block_frames = len(block_body["frame_exprs"])
            block_body = json.dumps(block_body).encode("utf-8")
            if compression == "gzip":
                block_body = gzip.compress(block_body, 1)
            block_body = base64.b64encode(block_body).decode("utf-8")
            req_blocks.append(
                {
                    "frames": block_frames,
                    "compression": compression,
                    "body": block_body,
                }
            )

        req = {
            "pos": pos,
            "terminal": terminal,
            "blocks": req_blocks,
        }
        response = self._session.post(
            f"{self._endpoint}/v2/spec/{spec_id}/part_block",
            json=req,
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response = response.json()
        assert response["status"] == "ok"

    def frame(self, width, height, pix_fmt, frame_expr, compression="gzip"):
        assert type(frame_expr) is FilterExpr or type(frame_expr) is SourceExpr
        assert compression is None or compression in ["gzip"]
        feb = _FrameExpressionBlock()
        feb.insert_frame(frame_expr)
        feb_body = feb.as_dict()

        feb_body = json.dumps(feb_body).encode("utf-8")
        if compression == "gzip":
            feb_body = gzip.compress(feb_body, 1)
        feb_body = base64.b64encode(feb_body).decode("utf-8")
        req = {
            "width": width,
            "height": height,
            "pix_fmt": pix_fmt,
            "compression": compression,
            "block": {
                "frames": 1,
                "compression": compression,
                "body": feb_body,
            },
        }
        response = self._session.post(
            f"{self._endpoint}/v2/frame",
            json=req,
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        if not response.ok:
            raise Exception(response.text)
        response_body = response.content
        assert type(response_body) is bytes
        if compression == "gzip":
            response_body = gzip.decompress(response_body)
        return response_body


class SourceExpr:
    def __init__(self, source, idx, is_iloc):
        self._source = source
        self._idx = idx
        self._is_iloc = is_iloc

    def __repr__(self):
        if self._is_iloc:
            return f"{self._source._name}.iloc[{self._idx}]"
        else:
            return f"{self._source._name}[{self._idx}]"

    def _to_json_spec(self):
        if self._is_iloc:
            return {
                "Source": {
                    "video": self._source._name,
                    "index": {"ILoc": int(self._idx)},
                }
            }
        else:
            return {
                "Source": {
                    "video": self._source._name,
                    "index": {"T": [self._idx.numerator, self._idx.denominator]},
                }
            }

    def _sources(self):
        return set([self._source])

    def _filters(self):
        return {}


class _SourceILoc:
    def __init__(self, source):
        self._source = source

    def __getitem__(self, idx):
        if type(idx) is not int:
            raise Exception(f"Source iloc index must be an integer, got a {type(idx)}")
        return SourceExpr(self._source, idx, True)


def _json_arg(arg, skip_data_anot=False):
    if type(arg) is FilterExpr or type(arg) is SourceExpr:
        return {"Frame": arg._to_json_spec()}
    elif type(arg) is int:
        if skip_data_anot:
            return {"Int": arg}
        return {"Data": {"Int": arg}}
    elif type(arg) is str:
        if skip_data_anot:
            return {"String": arg}
        return {"Data": {"String": arg}}
    elif type(arg) is bytes:
        arg = list(arg)
        if skip_data_anot:
            return {"Bytes": arg}
        return {"Data": {"Bytes": arg}}
    elif type(arg) is float:
        if skip_data_anot:
            return {"Float": arg}
        return {"Data": {"Float": arg}}
    elif type(arg) is bool:
        if skip_data_anot:
            return {"Bool": arg}
        return {"Data": {"Bool": arg}}
    elif type(arg) is tuple or type(arg) is list:
        if skip_data_anot:
            return {"List": [_json_arg(x, True) for x in list(arg)]}
        return {"Data": {"List": [_json_arg(x, True) for x in list(arg)]}}
    else:
        raise Exception(f"Unknown arg type: {type(arg)}")


class Filter:
    """A video filter."""

    def __init__(self, func: str):
        self._func = func

    def __call__(self, *args, **kwargs):
        return FilterExpr(self, args, kwargs)


class FilterExpr:
    def __init__(self, filter: Filter, args, kwargs):
        self._filter = filter
        self._args = args
        self._kwargs = kwargs

    def __repr__(self):
        args = []
        for arg in self._args:
            val = f'"{arg}"' if type(arg) is str else str(arg)
            args.append(str(val))
        for k, v in self._kwargs.items():
            val = f'"{v}"' if type(v) is str else str(v)
            args.append(f"{k}={val}")
        return f"{self._filter._func}({', '.join(args)})"

    def _to_json_spec(self):
        args = []
        for arg in self._args:
            args.append(_json_arg(arg))
        kwargs = {}
        for k, v in self._kwargs.items():
            kwargs[k] = _json_arg(v)
        return {"Filter": {"name": self._filter._func, "args": args, "kwargs": kwargs}}

    def _sources(self):
        s = set()
        for arg in self._args:
            if type(arg) is FilterExpr or type(arg) is SourceExpr:
                s = s.union(arg._sources())
        for arg in self._kwargs.values():
            if type(arg) is FilterExpr or type(arg) is SourceExpr:
                s = s.union(arg._sources())
        return s

    def _filters(self):
        f = {self._filter._func: self._filter}
        for arg in self._args:
            if type(arg) is FilterExpr:
                f = {**f, **arg._filters()}
        for arg in self._kwargs.values():
            if type(arg) is FilterExpr:
                f = {**f, **arg._filters()}
        return f



"""
vidformer.supervision is the [supervision](https://supervision.roboflow.com/) frontend for [vidformer](https://github.com/ixlab/vidformer).
"""

from math import sqrt

import numpy as np
import supervision as _sv
from supervision import Color, ColorLookup, ColorPalette, Detections
from supervision.annotators.utils import resolve_color, resolve_text_background_xyxy
from supervision.config import CLASS_NAME_DATA_FIELD

# supervision moved this between two versions, so we need to handle both cases
try:
    from supervision.detection.utils import spread_out_boxes
except ImportError:
    from supervision.detection.utils.boxes import spread_out_boxes

from supervision.geometry.core import Position

import vidformer.cv2 as vf_cv2

try:
    import cv2 as ocv_cv2
except ImportError:
    ocv_cv2 = None

CV2_FONT = vf_cv2.FONT_HERSHEY_SIMPLEX


class BoxAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        thickness=2,
        color_lookup=ColorLookup.CLASS,
    ):
        self.color = color
        self.thickness = thickness
        self.color_lookup = color_lookup

    def annotate(
        self,
        scene: vf_cv2.Frame,
        detections: Detections,
        custom_color_lookup=None,
    ):
        for detection_idx in range(len(detections)):
            x1, y1, x2, y2 = detections.xyxy[detection_idx].astype(int)
            color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=detection_idx,
                color_lookup=(
                    self.color_lookup
                    if custom_color_lookup is None
                    else custom_color_lookup
                ),
            )
            vf_cv2.rectangle(
                img=scene,
                pt1=(x1, y1),
                pt2=(x2, y2),
                color=color.as_bgr(),
                thickness=self.thickness,
            )
        return scene


class RoundBoxAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        thickness: int = 2,
        color_lookup: ColorLookup = ColorLookup.CLASS,
        roundness: float = 0.6,
    ):
        self.color = color
        self.thickness = thickness
        self.color_lookup = color_lookup
        if not 0 < roundness <= 1.0:
            raise ValueError("roundness attribute must be float between (0, 1.0]")
        self.roundness = roundness

    def annotate(
        self,
        scene: vf_cv2.Frame,
        detections: _sv.Detections,
        custom_color_lookup=None,
    ):
        for detection_idx in range(len(detections)):
            x1, y1, x2, y2 = detections.xyxy[detection_idx].astype(int)
            color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=detection_idx,
                color_lookup=(
                    self.color_lookup
                    if custom_color_lookup is None
                    else custom_color_lookup
                ),
            )
            radius = (
                int((x2 - x1) // 2 * self.roundness)
                if abs(x1 - x2) < abs(y1 - y2)
                else int((y2 - y1) // 2 * self.roundness)
            )
            circle_coordinates = [
                ((x1 + radius), (y1 + radius)),
                ((x2 - radius), (y1 + radius)),
                ((x2 - radius), (y2 - radius)),
                ((x1 + radius), (y2 - radius)),
            ]
            line_coordinates = [
                ((x1 + radius, y1), (x2 - radius, y1)),
                ((x2, y1 + radius), (x2, y2 - radius)),
                ((x1 + radius, y2), (x2 - radius, y2)),
                ((x1, y1 + radius), (x1, y2 - radius)),
            ]
            start_angles = (180, 270, 0, 90)
            end_angles = (270, 360, 90, 180)
            for center_coordinates, line, start_angle, end_angle in zip(
                circle_coordinates, line_coordinates, start_angles, end_angles
            ):
                vf_cv2.ellipse(
                    img=scene,
                    center=center_coordinates,
                    axes=(radius, radius),
                    angle=0,
                    startAngle=start_angle,
                    endAngle=end_angle,
                    color=color.as_bgr(),
                    thickness=self.thickness,
                )
                vf_cv2.line(
                    img=scene,
                    pt1=line[0],
                    pt2=line[1],
                    color=color.as_bgr(),
                    thickness=self.thickness,
                )
        return scene


class BoxCornerAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        thickness=4,
        corner_length=15,
        color_lookup=ColorLookup.CLASS,
    ):
        self.color = color
        self.thickness: int = thickness
        self.corner_length: int = corner_length
        self.color_lookup: ColorLookup = color_lookup

    def annotate(
        self,
        scene: vf_cv2.Frame,
        detections: Detections,
        custom_color_lookup=None,
    ):
        for detection_idx in range(len(detections)):
            x1, y1, x2, y2 = detections.xyxy[detection_idx].astype(int)
            color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=detection_idx,
                color_lookup=(
                    self.color_lookup
                    if custom_color_lookup is None
                    else custom_color_lookup
                ),
            )
            corners = [(x1, y1), (x2, y1), (x1, y2), (x2, y2)]
            for x, y in corners:
                x_end = x + self.corner_length if x == x1 else x - self.corner_length
                vf_cv2.line(
                    scene, (x, y), (x_end, y), color.as_bgr(), thickness=self.thickness
                )

                y_end = y + self.corner_length if y == y1 else y - self.corner_length
                vf_cv2.line(
                    scene, (x, y), (x, y_end), color.as_bgr(), thickness=self.thickness
                )
        return scene


class ColorAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        opacity: float = 0.5,
        color_lookup: ColorLookup = ColorLookup.CLASS,
    ):
        self.color = color
        self.color_lookup: ColorLookup = color_lookup
        self.opacity = opacity

    def annotate(
        self,
        scene: vf_cv2.Frame,
        detections: Detections,
        custom_color_lookup=None,
    ):
        scene_with_boxes = scene.copy()
        for detection_idx in range(len(detections)):
            x1, y1, x2, y2 = detections.xyxy[detection_idx].astype(int)
            color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=detection_idx,
                color_lookup=(
                    self.color_lookup
                    if custom_color_lookup is None
                    else custom_color_lookup
                ),
            )
            vf_cv2.rectangle(
                img=scene_with_boxes,
                pt1=(x1, y1),
                pt2=(x2, y2),
                color=color.as_bgr(),
                thickness=-1,
            )

        vf_cv2.addWeighted(
            scene_with_boxes, self.opacity, scene, 1 - self.opacity, gamma=0, dst=scene
        )
        return scene


class CircleAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        thickness: int = 2,
        color_lookup: ColorLookup = ColorLookup.CLASS,
    ):
        self.color = color
        self.thickness: int = thickness
        self.color_lookup: ColorLookup = color_lookup

    def annotate(
        self,
        scene: vf_cv2.Frame,
        detections: Detections,
        custom_color_lookup=None,
    ):
        for detection_idx in range(len(detections)):
            x1, y1, x2, y2 = detections.xyxy[detection_idx].astype(int)
            center = ((x1 + x2) // 2, (y1 + y2) // 2)
            distance = sqrt((x1 - center[0]) ** 2 + (y1 - center[1]) ** 2)
            color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=detection_idx,
                color_lookup=(
                    self.color_lookup
                    if custom_color_lookup is None
                    else custom_color_lookup
                ),
            )
            vf_cv2.circle(
                img=scene,
                center=center,
                radius=int(distance),
                color=color.as_bgr(),
                thickness=self.thickness,
            )

        return scene


class DotAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        radius: int = 4,
        position: Position = Position.CENTER,
        color_lookup: ColorLookup = ColorLookup.CLASS,
        outline_thickness: int = 0,
        outline_color=Color.BLACK,
    ):
        self.color = color
        self.radius: int = radius
        self.position: Position = position
        self.color_lookup: ColorLookup = color_lookup
        self.outline_thickness = outline_thickness
        self.outline_color = outline_color

    def annotate(
        self,
        scene: vf_cv2.Frame,
        detections: Detections,
        custom_color_lookup=None,
    ):
        xy = detections.get_anchors_coordinates(anchor=self.position)
        for detection_idx in range(len(detections)):
            color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=detection_idx,
                color_lookup=(
                    self.color_lookup
                    if custom_color_lookup is None
                    else custom_color_lookup
                ),
            )
            center = (int(xy[detection_idx, 0]), int(xy[detection_idx, 1]))

            vf_cv2.circle(scene, center, self.radius, color.as_bgr(), -1)
            if self.outline_thickness:
                outline_color = resolve_color(
                    color=self.outline_color,
                    detections=detections,
                    detection_idx=detection_idx,
                    color_lookup=(
                        self.color_lookup
                        if custom_color_lookup is None
                        else custom_color_lookup
                    ),
                )
                vf_cv2.circle(
                    scene,
                    center,
                    self.radius,
                    outline_color.as_bgr(),
                    self.outline_thickness,
                )
        return scene


class LabelAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        text_color=Color.WHITE,
        text_scale: float = 0.5,
        text_thickness: int = 1,
        text_padding: int = 10,
        text_position: Position = Position.TOP_LEFT,
        color_lookup: ColorLookup = ColorLookup.CLASS,
        border_radius: int = 0,
        smart_position: bool = False,
    ):
        self.border_radius: int = border_radius
        self.color = color
        self.text_color = text_color
        self.text_scale: float = text_scale
        self.text_thickness: int = text_thickness
        self.text_padding: int = text_padding
        self.text_anchor: Position = text_position
        self.color_lookup: ColorLookup = color_lookup
        self.smart_position = smart_position

    def annotate(
        self,
        scene,
        detections: Detections,
        labels,
        custom_color_lookup=None,
    ):
        self._validate_labels(labels, detections)

        labels = self._get_labels_text(detections, labels)
        label_properties = self._get_label_properties(detections, labels)

        if self.smart_position:
            xyxy = label_properties[:, :4]
            xyxy = spread_out_boxes(xyxy)
            label_properties[:, :4] = xyxy

        self._draw_labels(
            scene=scene,
            labels=labels,
            label_properties=label_properties,
            detections=detections,
            custom_color_lookup=custom_color_lookup,
        )

        return scene

    def _validate_labels(self, labels, detections: Detections):
        if labels is not None and len(labels) != len(detections):
            raise ValueError(
                f"The number of labels ({len(labels)}) does not match the "
                f"number of detections ({len(detections)}). Each detection "
                f"should have exactly 1 label."
            )

    def _get_label_properties(
        self,
        detections: Detections,
        labels,
    ):
        label_properties = []
        anchors_coordinates = detections.get_anchors_coordinates(
            anchor=self.text_anchor
        ).astype(int)

        for label, center_coords in zip(labels, anchors_coordinates):
            text_w, text_h = vf_cv2.getTextSize(
                text=label,
                fontFace=CV2_FONT,
                fontScale=self.text_scale,
                thickness=self.text_thickness,
            )[0]

            width_padded = text_w + 2 * self.text_padding
            height_padded = text_h + 2 * self.text_padding

            text_background_xyxy = resolve_text_background_xyxy(
                center_coordinates=tuple(center_coords),
                text_wh=(width_padded, height_padded),
                position=self.text_anchor,
            )

            label_properties.append(
                [
                    *text_background_xyxy,
                    text_h,
                ]
            )

        return np.array(label_properties).reshape(-1, 5)

    @staticmethod
    def _get_labels_text(detections: Detections, custom_labels):
        if custom_labels is not None:
            return custom_labels

        labels = []
        for idx in range(len(detections)):
            if CLASS_NAME_DATA_FIELD in detections.data:
                labels.append(detections.data[CLASS_NAME_DATA_FIELD][idx])
            elif detections.class_id is not None:
                labels.append(str(detections.class_id[idx]))
            else:
                labels.append(str(idx))
        return labels

    def _draw_labels(
        self,
        scene,
        labels,
        label_properties,
        detections,
        custom_color_lookup,
    ) -> None:
        assert len(labels) == len(label_properties) == len(detections), (
            f"Number of label properties ({len(label_properties)}), "
            f"labels ({len(labels)}) and detections ({len(detections)}) "
            "do not match."
        )

        color_lookup = (
            custom_color_lookup
            if custom_color_lookup is not None
            else self.color_lookup
        )

        for idx, label_property in enumerate(label_properties):
            background_color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=idx,
                color_lookup=color_lookup,
            )
            text_color = resolve_color(
                color=self.text_color,
                detections=detections,
                detection_idx=idx,
                color_lookup=color_lookup,
            )

            box_xyxy = label_property[:4]
            text_height_padded = label_property[4]
            self.draw_rounded_rectangle(
                scene=scene,
                xyxy=box_xyxy,
                color=background_color.as_bgr(),
                border_radius=self.border_radius,
            )

            text_x = box_xyxy[0] + self.text_padding
            text_y = box_xyxy[1] + self.text_padding + text_height_padded
            vf_cv2.putText(
                img=scene,
                text=labels[idx],
                org=(text_x, text_y),
                fontFace=CV2_FONT,
                fontScale=self.text_scale,
                color=text_color.as_bgr(),
                thickness=self.text_thickness,
                lineType=vf_cv2.LINE_AA,
            )

    @staticmethod
    def draw_rounded_rectangle(
        scene: np.ndarray,
        xyxy,
        color,
        border_radius: int,
    ) -> np.ndarray:
        x1, y1, x2, y2 = xyxy
        width = x2 - x1
        height = y2 - y1

        border_radius = min(border_radius, min(width, height) // 2)

        if border_radius <= 0:
            vf_cv2.rectangle(
                img=scene,
                pt1=(x1, y1),
                pt2=(x2, y2),
                color=color,
                thickness=-1,
            )
        else:
            rectangle_coordinates = [
                ((x1 + border_radius, y1), (x2 - border_radius, y2)),
                ((x1, y1 + border_radius), (x2, y2 - border_radius)),
            ]
            circle_centers = [
                (x1 + border_radius, y1 + border_radius),
                (x2 - border_radius, y1 + border_radius),
                (x1 + border_radius, y2 - border_radius),
                (x2 - border_radius, y2 - border_radius),
            ]

            for coordinates in rectangle_coordinates:
                vf_cv2.rectangle(
                    img=scene,
                    pt1=coordinates[0],
                    pt2=coordinates[1],
                    color=color,
                    thickness=-1,
                )
            for center in circle_centers:
                vf_cv2.circle(
                    img=scene,
                    center=center,
                    radius=border_radius,
                    color=color,
                    thickness=-1,
                )
        return scene


class MaskAnnotator:
    def __init__(
        self,
        color=ColorPalette.DEFAULT,
        opacity: float = 0.5,
        color_lookup: ColorLookup = ColorLookup.CLASS,
    ):
        self.color = color
        self.opacity = opacity
        self.color_lookup: ColorLookup = color_lookup

    def annotate(
        self,
        scene,
        detections: Detections,
        custom_color_lookup=None,
    ):
        if detections.mask is None:
            return scene

        colored_mask = scene.copy()

        for detection_idx in np.flip(np.argsort(detections.box_area)):
            color = resolve_color(
                color=self.color,
                detections=detections,
                detection_idx=detection_idx,
                color_lookup=(
                    self.color_lookup
                    if custom_color_lookup is None
                    else custom_color_lookup
                ),
            )
            mask = detections.mask[detection_idx]
            colored_mask[mask] = color.as_bgr()

        vf_cv2.addWeighted(
            colored_mask, self.opacity, scene, 1 - self.opacity, 0, dst=scene
        )
        return scene


class MaskStreamWriter:
    def __init__(self, path: str, shape: tuple):
        # Shape should be (width, height)
        assert ocv_cv2 is not None, "OpenCV cv2 is required for ExternDetectionsBuilder"
        assert type(shape) is tuple, "shape must be a tuple"
        assert len(shape) == 2, "shape must be a tuple of length 2"
        self._shape = (shape[1], shape[0])
        self._writer = ocv_cv2.VideoWriter(
            path, ocv_cv2.VideoWriter_fourcc(*"FFV1"), 1, shape, isColor=False
        )
        assert self._writer.isOpened(), f"Failed to open video writer at {path}"
        self._i = 0

    def write_detections(self, detections: Detections):
        if len(detections) == 0:
            return self._i

        mask = detections.mask
        assert (
            mask.shape[1:] == self._shape
        ), f"mask shape ({mask.shape[1:]}) must match the shape of the video ({self._shape})"
        for i in range(mask.shape[0]):
            frame_uint8 = detections.mask[i].astype(np.uint8)
            self._writer.write(frame_uint8)
            self._i += 1
        return self._i

    def release(self):
        self._writer.release()


def populate_mask(
    detections: Detections, mask_stream: vf_cv2.VideoCapture, frame_idx: int
):
    assert type(detections) is Detections
    assert detections.mask is None
    detections.mask = []
    assert len(detections) + frame_idx <= len(mask_stream)
    for i in range(len(detections)):
        mask = mask_stream[frame_idx + i]
        assert mask.shape[2] == 1, "mask must be a single channel image"
        detections.mask.append(mask)



"""
vidformer.numpy mimics numpy's API while supporting vidformer Frames.

When Frames are detected, dispatches to cv2 functions. Otherwise falls back to numpy.
"""

import numpy as _np
import vidformer.cv2 as _vf_cv2

# Re-export numpy's uint8 and other common dtypes
uint8 = _np.uint8
float32 = _np.float32
float64 = _np.float64
int32 = _np.int32
int64 = _np.int64


def _is_frame(obj):
    return isinstance(obj, _vf_cv2.Frame)


def _contains_frame(iterable):
    return any(_is_frame(x) for x in iterable)


def _is_frame_shape(shape, dtype):
    """Check if shape/dtype combination should produce a Frame."""
    return (
        isinstance(shape, (tuple, list))
        and len(shape) == 3
        and shape[2] in (1, 3)
        and (dtype is None or dtype == _np.uint8)
    )


def zeros(shape, dtype=None, *args, **kwargs):
    """Mimics numpy.zeros. Returns a Frame for video-like shapes (h, w, 1|3) with uint8."""
    if not args and not kwargs and _is_frame_shape(shape, dtype):
        return _vf_cv2.zeros(shape, dtype=_np.uint8)

    if dtype is None:
        return _np.zeros(shape, *args, **kwargs)
    return _np.zeros(shape, dtype=dtype, *args, **kwargs)


def ones(shape, dtype=None, *args, **kwargs):
    """Mimics numpy.ones. Returns a Frame for video-like shapes (h, w, 1|3) with uint8."""
    if not args and not kwargs and _is_frame_shape(shape, dtype):
        return _vf_cv2.solid(shape, (1, 1, 1), dtype=_np.uint8)

    if dtype is None:
        return _np.ones(shape, *args, **kwargs)
    return _np.ones(shape, dtype=dtype, *args, **kwargs)


def _parse_fill_value(fill_value):
    """Convert fill_value to a color tuple, or return None if unsupported."""
    if isinstance(fill_value, (int, float)):
        v = int(fill_value)
        return (v, v, v)
    elif isinstance(fill_value, (tuple, list)) and len(fill_value) == 3:
        return tuple(int(c) for c in fill_value)
    return None


def full(shape, fill_value, dtype=None, *args, **kwargs):
    """Mimics numpy.full. Returns a Frame for video-like shapes (h, w, 1|3) with uint8."""
    if not args and not kwargs and _is_frame_shape(shape, dtype):
        color = _parse_fill_value(fill_value)
        if color is not None:
            return _vf_cv2.solid(shape, color, dtype=_np.uint8)

    if dtype is None:
        return _np.full(shape, fill_value, *args, **kwargs)
    return _np.full(shape, fill_value, dtype=dtype, *args, **kwargs)


def zeros_like(a, dtype=None, *args, **kwargs):
    """Mimics numpy.zeros_like. Returns a Frame if input is a Frame."""
    if _is_frame(a):
        return _vf_cv2.zeros(a.shape, dtype=_np.uint8)
    return _np.zeros_like(a, dtype=dtype, *args, **kwargs)


def ones_like(a, dtype=None, *args, **kwargs):
    """Mimics numpy.ones_like. Returns a Frame if input is a Frame."""
    if _is_frame(a):
        return _vf_cv2.solid(a.shape, (1, 1, 1), dtype=_np.uint8)
    return _np.ones_like(a, dtype=dtype, *args, **kwargs)


def full_like(a, fill_value, dtype=None, *args, **kwargs):
    """Mimics numpy.full_like. Returns a Frame if input is a Frame."""
    if _is_frame(a):
        color = _parse_fill_value(fill_value)
        if color is not None:
            return _vf_cv2.solid(a.shape, color, dtype=_np.uint8)
        # Fall back to numpy for unsupported fill_value types
        return _np.full_like(a.numpy(), fill_value, dtype=dtype, *args, **kwargs)
    return _np.full_like(a, fill_value, dtype=dtype, *args, **kwargs)


def hstack(tup, *args, **kwargs):
    """Mimics numpy.hstack. Uses cv2.hconcat for Frames."""
    if _contains_frame(tup):
        return _vf_cv2.hconcat(tup)
    return _np.hstack(tup, *args, **kwargs)


def vstack(tup, *args, **kwargs):
    """Mimics numpy.vstack. Uses cv2.vconcat for Frames."""
    if _contains_frame(tup):
        return _vf_cv2.vconcat(tup)
    return _np.vstack(tup, *args, **kwargs)


def concatenate(arrays, axis=0, *args, **kwargs):
    """Mimics numpy.concatenate. Uses cv2.hconcat/vconcat for Frames."""
    if _contains_frame(arrays):
        if axis == 0:
            return _vf_cv2.vconcat(arrays)
        elif axis == 1:
            return _vf_cv2.hconcat(arrays)
        else:
            raise ValueError(
                f"Unsupported axis {axis} for Frame concatenation (only 0 or 1)"
            )
    return _np.concatenate(arrays, axis=axis, *args, **kwargs)


def flip(m, axis=None, *args, **kwargs):
    """Mimics numpy.flip. Uses cv2.flip for Frames."""
    if _is_frame(m):
        if axis is None:
            return _vf_cv2.flip(m, -1)
        elif axis == 0:
            return _vf_cv2.flip(m, 0)
        elif axis == 1:
            return _vf_cv2.flip(m, 1)
        else:
            raise ValueError(
                f"Unsupported axis {axis} for Frame flip (only 0, 1, or None)"
            )
    return _np.flip(m, axis=axis, *args, **kwargs)


def rot90(m, k=1, axes=(0, 1), *args, **kwargs):
    """Mimics numpy.rot90. Uses cv2.rotate for Frames."""
    if _is_frame(m):
        if axes != (0, 1):
            raise ValueError("Only axes=(0, 1) supported for Frame rotation")
        k = k % 4
        if k == 0:
            return m
        elif k == 1:
            return _vf_cv2.rotate(m, _vf_cv2.ROTATE_90_COUNTERCLOCKWISE)
        elif k == 2:
            return _vf_cv2.rotate(m, _vf_cv2.ROTATE_180)
        elif k == 3:
            return _vf_cv2.rotate(m, _vf_cv2.ROTATE_90_CLOCKWISE)
    return _np.rot90(m, k=k, axes=axes, *args, **kwargs)


def __getattr__(name):
    """Passthrough for all other numpy attributes."""
    return getattr(_np, name)
